"use strict";

var _interopRequireWildcard = require("@babel/runtime/helpers/interopRequireWildcard");

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = loader;
exports.generatePaths = generatePaths;

var _regenerator = _interopRequireDefault(require("@babel/runtime/regenerator"));

var _asyncToGenerator2 = _interopRequireDefault(require("@babel/runtime/helpers/asyncToGenerator"));

var _defineProperty2 = _interopRequireDefault(require("@babel/runtime/helpers/defineProperty"));

var _objectSpread3 = _interopRequireDefault(require("@babel/runtime/helpers/objectSpread"));

var _toConsumableArray2 = _interopRequireDefault(require("@babel/runtime/helpers/toConsumableArray"));

var babylon = _interopRequireWildcard(require("babylon"));

var _babelTraverse = _interopRequireDefault(require("babel-traverse"));

var _babelGenerator = _interopRequireDefault(require("babel-generator"));

var t = _interopRequireWildcard(require("babel-types"));

var _babelTemplate = _interopRequireDefault(require("babel-template"));

var _utils = require("./utils");

var UNEXISTING = 'UNEXISTING';
var buildDynamicImport = (0, _babelTemplate.default)("\n  new Promise((resolve) => {\n    require.ensure([], (require) => {\n      resolve(require(PACKAGE_NAME));\n    });\n  })\n");
var buildRequire = (0, _babelTemplate.default)("\n  require(DEF_FILE).default ? require(DEF_FILE).default : require(DEF_FILE)\n");

function loader(source, map, meta) {
  var ast = babylon.parse(source);
  var that = this;
  var callback = this.async();
  var promises = []; // first traverse, we collect all resolve promise and store in promises array

  traverseAst(ast, function (_ref) {
    var packageName = _ref.packageName;
    promises.push(generatePaths(packageName, that));
  }, function (_ref2) {
    var packageName = _ref2.packageName;
    promises.push(generatePaths(packageName, that));
  });
  (0, _babelGenerator.default)(ast, {}, source).code; // second traverse, we get the generatePaths from the promises array

  Promise.all(promises).then(function (paths) {
    var copyPaths = paths.slice();
    traverseAst(ast, function (_ref3) {
      var path = _ref3.path,
          type = _ref3.type;

      var _copyPaths$shift = copyPaths.shift(),
          absPackageName = _copyPaths$shift.absPackageName,
          absPackageJson = _copyPaths$shift.absPackageJson,
          absDefJs = _copyPaths$shift.absDefJs,
          packageName = _copyPaths$shift.packageName;

      path.get('arguments.1').replaceWith(t.objectExpression([t.objectProperty(t.stringLiteral('packageName'), t.stringLiteral(packageName)), t.objectProperty(t.stringLiteral('loader'), buildLoader(absPackageName)), t.objectProperty(t.stringLiteral('builder'), buildBuilder(absDefJs, that))].concat((0, _toConsumableArray2.default)(buildCannerConfig(absPackageJson, type)))));
    }, function (_ref4) {
      var propsNode = _ref4.propsNode,
          type = _ref4.type;

      var _copyPaths$shift2 = copyPaths.shift(),
          absPackageName = _copyPaths$shift2.absPackageName,
          absPackageJson = _copyPaths$shift2.absPackageJson,
          absDefJs = _copyPaths$shift2.absDefJs,
          packageName = _copyPaths$shift2.packageName;

      propsNode.node.properties.push(t.objectProperty(t.stringLiteral('packageName'), t.stringLiteral(packageName)));
      propsNode.node.properties.push(t.objectProperty(t.stringLiteral('loader'), buildLoader(absPackageName)));
      propsNode.node.properties.push(t.objectProperty(t.stringLiteral('builder'), buildBuilder(absDefJs, that)));
      propsNode.node.properties = propsNode.node.properties.concat(buildCannerConfig(absPackageJson, type));
    });
    var result = (0, _babelGenerator.default)(ast, {}, source).code;
    callback(null, result, map, meta);
  });
}

function traverseAst(ast, nullPropsCallback, ObjectPropsCallback) {
  (0, _babelTraverse.default)(ast, {
    CallExpression: function CallExpression(path) {
      var typeNode, propsNode, type, packageName;

      if (path.node.arguments.length < 2) {
        return;
      }

      typeNode = path.get('arguments.0');
      propsNode = path.get('arguments.1');

      if (typeNode.isStringLiteral() && propsNode.isNullLiteral()) {
        type = typeNode.node.value;
        packageName = _utils.componentMap.get('type');

        if (!packageName) {
          return;
        }

        nullPropsCallback({
          packageName: packageName,
          type: type,
          propsNode: propsNode,
          path: path
        });
      }

      if (typeNode.isStringLiteral() && propsNode.isObjectExpression()) {
        type = typeNode.node.value;
        var properties = getProperties(propsNode);

        if ('packageName' in properties) {
          packageName = properties.packageName;
        } else {
          packageName = _utils.componentMap.get(type, properties.ui);
        }

        if (!packageName) {
          return;
        }

        ObjectPropsCallback({
          packageName: packageName,
          type: type,
          path: path,
          propsNode: propsNode
        });
      }
    }
  });
}

function getProperties(propsNode) {
  return propsNode.node.properties.reduce(function (result, property) {
    return (0, _objectSpread3.default)({}, result, (0, _defineProperty2.default)({}, getPropertyKey(property), getPropertyValue(property)));
  }, {});
}

function getPropertyKey(property) {
  return property.key.name;
}

function getPropertyValue(property) {
  if (t.isStringLiteral(property.value)) {
    return property.value.value;
  } // for now, we only need ui and packagename, so If it's not string literal we can ignore it.


  return undefined;
}

function absPath(path, context) {
  return new Promise(function (resolve) {
    context.resolve(context.context, path, function (err, result) {
      if (err) {
        resolve(UNEXISTING);
      }

      resolve(result);
    });
  });
}

function generatePaths(_x, _x2) {
  return _generatePaths.apply(this, arguments);
}

function _generatePaths() {
  _generatePaths = (0, _asyncToGenerator2.default)(
  /*#__PURE__*/
  _regenerator.default.mark(function _callee(sourcePkgName, context) {
    var absPackageName, absPackageJson, absDefJs;
    return _regenerator.default.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            absPackageName = sourcePkgName;
            absPackageJson = "".concat(sourcePkgName, "/package.json");
            absDefJs = "".concat(sourcePkgName, "/canner.def.js");
            _context.next = 5;
            return absPath(absPackageName, context);

          case 5:
            absPackageName = _context.sent;
            _context.next = 8;
            return absPath(absPackageJson, context);

          case 8:
            absPackageJson = _context.sent;
            _context.next = 11;
            return absPath(absDefJs, context);

          case 11:
            absDefJs = _context.sent;
            return _context.abrupt("return", {
              absPackageName: absPackageName,
              absPackageJson: absPackageJson,
              absDefJs: absDefJs,
              packageName: sourcePkgName
            });

          case 13:
          case "end":
            return _context.stop();
        }
      }
    }, _callee, this);
  }));
  return _generatePaths.apply(this, arguments);
}

function buildLoader(packageName) {
  return buildDynamicImport({
    PACKAGE_NAME: t.stringLiteral(packageName)
  }).expression;
}

function buildBuilder(defFile) {
  if (defFile === UNEXISTING) {
    return t.nullLiteral();
  }

  return buildRequire({
    DEF_FILE: t.stringLiteral(defFile)
  }).expression;
}

function buildCannerConfig(packageJson, type) {
  var canner = {},
      properties = [];

  try {
    canner = require(packageJson).canner || {};
  } catch (e) {
    canner = {};
  }

  Object.keys(canner).forEach(function (key) {
    if (key === 'cannerDataType') {
      if (type !== 'json') {
        // the component of json type is same as object, but the type can't be changed
        properties.push(t.objectProperty(t.stringLiteral('type'), t.stringLiteral(canner[key])));
      }
    } else if (typeof canner[key] === 'string') {
      properties.push(t.objectProperty(t.stringLiteral(key), t.stringLiteral(canner[key])));
    } else if (typeof canner[key] === 'number') {
      properties.push(t.objectProperty(t.stringLiteral(key), t.numberLiteral(canner[key])));
    } else if (typeof canner[key] === 'boolean') {
      properties.push(t.objectProperty(t.stringLiteral(key), t.booleanLiteral(canner[key])));
    } else {
      throw new Error("Not support the type of object and array config in canner yet, but got key: ".concat(key, " with value: ").concat(canner[key], "."));
    }
  });
  return properties;
}